<?php
/**
 * Perforce Swarm
 *
 * @copyright   2013-2020 Perforce Software. All rights reserved.
 * @license     Please see LICENSE.txt in top-level readme folder of this distribution.
 * @version     2019.3/2027974
 */

namespace Workflow\Listener;

use Activity\Model\Activity;
use Application\Model\IModelDAO;
use Events\Listener\AbstractEventListener;
use Projects\Model\Project as ProjectModel;
use Zend\EventManager\Event;

/**
 * Listener class to handle workflow events
 * @package Workflow\Listener
 */
class WorkflowListener extends AbstractEventListener
{
    const CREATED_ACTION = 'created';
    const UPDATED_ACTION = 'updated';

    /**
     * Handle workflow created
     * @param Event $event the event
     */
    public function workflowCreated(Event $event)
    {
        $this->createActivity($event);
    }

    /**
     * Handle workflow updated
     * @param Event $event the event
     */
    public function workflowUpdated(Event $event)
    {
        $this->createActivity($event, self::UPDATED_ACTION);
        $this->linkToProjects($event);
    }

    /**
     * Updates the existing activity created on the event to add a project link when required.
     * A project link will describe the projects/branches linked to the workflow on the activity
     * and is only required when a workflow is updated.
     * @param Event $event  the event
     */
    private function linkToProjects(Event $event)
    {
        try {
            $activity     = $event->getParam('activity');
            $wfId         = $event->getParam('id');
            $p4admin      = $this->services->get('p4_admin');
            $projectDAO   = $this->services->get(IModelDAO::PROJECT_DAO);
            $projects     = $projectDAO->fetchAll([ProjectModel::FETCH_BY_WORKFLOW => $wfId], $p4admin);
            $projectsLink = [];
            foreach ($projects as $project) {
                $projectWf                = $project->getWorkflow();
                $projectId                = $project->getId();
                $projectsLink[$projectId] = [];
                $branches                 = $project->getBranches();
                foreach ($branches as $branch) {
                    $branchWf = $project->getWorkflow($branch['id']);
                    if ($branchWf && $branchWf != $projectWf) {
                        $projectsLink[$projectId][] = $branch['id'];
                    }
                }
            }
            $activity->set(
                ['projects' => $projectsLink]
            );
        } catch (\Exception $e) {
            $this->services->get('logger')->err($e->getMessage());
        }
    }

    /**
     * Create activity when a workflow is created or updated.
     * @param Event     $event      the event
     * @param string    $action     the action, defaults to 'created'
     */
    private function createActivity(Event $event, $action = self::CREATED_ACTION)
    {
        try {
            $data     = $event->getParam('data');
            $p4admin  = $this->services->get('p4_admin');
            $wf       = $this->services->get(IModelDAO::WORKFLOW_DAO)->fetch($event->getParam('id'), $p4admin);
            $activity = new Activity;
            $activity->set(
                [
                    'action'   => $action,
                    'user'     => $data['user'],
                    'target'   => 'workflow (' . $wf->getName() . ')'
                ]
            );
            $event->setParam('activity', $activity);
        } catch (\Exception $e) {
            $this->services->get('logger')->err($e->getMessage());
        }
    }
}
